/*****************************************************************************
 * 
 * Sampled sound tempo changer. Changes the sound tempo while maintaining the
 * original pitch by using time domain harmonic scaling (TDHS) method with 
 * several performance-increasing tweaks.
 *
 * Note : MMX optimized functions reside in a separate, platform-specific file, 
 * e.g. 'mmx_win.cpp' or 'mmx_gcc.cpp'
 *
 * Author        : Copyright (c) Olli Parviainen
 * Author e-mail : oparviai @ iki.fi
 * File created  : 13-Jan-2002
 *
 * Last changed  : $Date: 2003/09/21 10:50:39 $
 * File revision : $Revision: 1.4 $
 *
 * $Id: TDStretch.h,v 1.4 2003/09/21 10:50:39 Olli Exp $
 *
 * License :
 * 
 *  SoundTouch sound processing library
 *  Copyright (c) Olli Parviainen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#ifndef TDStretch_H
#define TDStretch_H

#include "STTypes.h"
#include "RateTransposer.h"
#include "FIFOSamplePipe.h"


class TDStretch : public FIFOProcessor
{
protected:
    uint channels;
    uint sampleReq;
    float tempo;

    short *pMidBuffer;
    short *pRefMidBuffer;
    uint overlapLength;
    uint overlapDividerBits;
    uint slopingDividerBits;
    uint seekLength;
    uint seekWindowLength;
    uint maxOffset;
    FIFOSampleBuffer outputBuffer;
    FIFOSampleBuffer inputBuffer;
    BOOL bQuickseek;
    BOOL bMidBufferDirty;

    virtual uint seekBestOverlapPositionStereo(const short *refPos);
    virtual uint seekBestOverlapPositionStereoQuick(const short *refPos);
    virtual uint seekBestOverlapPositionMono(const short *refPos);
    virtual uint seekBestOverlapPositionMonoQuick(const short *refPos);
    uint seekBestOverlapPosition(const short *refPos);

    virtual void overlapStereo(short *output, const short *input) const;
    virtual void overlapMono(short *output, const short *input) const;

    void clearMidBuffer();
    void overlap(short *output, const short *input, const uint ovlPos) const;

    void slopeReferenceSamplesStereo();

    void processNominalTempo();

    // Changes the tempo of the given sound samples.
    // Returns amount of samples returned in the "output" buffer.
    // The maximum amount of samples that can be returned at a time is set by
    // the 'set_returnBuffer_size' function.
    void processSamples();
    
public:
    TDStretch();
    virtual ~TDStretch();

    // Operator 'new' is overloaded so that it automatically creates a suitable instance 
    // depending on if we've a MMX-capable CPU available or not.
    void *operator new(size_t s);

    static TDStretch *newInstance();
    
    // Returns the output buffer object
    FIFOSamplePipe *getOutput() { return &outputBuffer; };

    // Returns the input buffer object
    FIFOSamplePipe *getInput() { return &inputBuffer; };

    // Sets new target tempo. Normal tempo = 'SCALE', smaller values represent slower 
    // tempo, larger faster tempo.
    void setTempo(const float newTempo);

    // Returns nonzero if there aren't any samples available for outputting.
    virtual void clear();

    // Clears the input buffer
    void clearInput();

    // Sets the number of channels, 1 = mono, 2 = stereo
    void setChannels(const uint numChannels);

    // Enables/disables the quick position seeking algorithm. Zero to disable, 
    // nonzero to enable
    void enableQuickSeek(const BOOL enable);

    // Returns nonzero if the quick seeking algorithm is enabled.
    BOOL isQuickSeekEnabled() const;

    // Sets routine control parameters. These control are certain time constants
    // defining how the sound is stretched to the desired duration.
    //
    // 'sampleRate' = sample rate of the sound
    // 'sequenceMS' = one processing sequence length in milliseconds (default = 82 ms)
    // 'seekwindowMS' = seeking window length for scanning the best overlapping 
    //      position (default = 28 ms)
    // 'overlapMS' = overlapping length (default = 12 ms)
    void setParameters(const uint sampleRate, const uint sequenceMS = 82, 
        const uint seekwindowMS = 28, const uint overlapMS = 12);

    // Adds 'numsamples' pcs of samples from the 'samples' memory position into
    // the input of the object.
    virtual void putSamples(const short *samples, const uint numSamples);
};


// MMX-optimized version of TDStretch class: The MMX-optimized routines are 
// overloaded in this version.
class TDStretchMMX : public TDStretch
{
private:
    int calculateCrossCorrelation(const short *mixingPos, const short *compare) const;

protected:
    virtual uint seekBestOverlapPositionStereoQuick(const short *other);
    virtual uint seekBestOverlapPositionStereo(const short *other);

    virtual void overlapStereo(short *output, const short *input) const;
};

#endif
